open Unix

type system_error =
| EIO       (* Erreur de bas niveau lors d'un accès au disque *)
| EACCESS   (* Erreur d'accès *)
| EISDIR    (* Répertoire trouvé où un fichier régulier est attendu *)
| ENOTDIR   (* Fichier régulier trouvé où un répertoire est attendu *)
| EINVAL    (* Argument invalide *)
| EEXIST    (* Un fichier existe déjà *)
| ENOENT    (* Un fichier manque *)
| EBUSY     (* Un fichier est occupé *)
| ENOSPC    (* Plus d'espace sur le disque *)

exception System of system_error * string * string

let system_error err call mes = raise (System (err, call, mes))

let handle_unix f x =
  try f x with Unix.Unix_error (error,s,mes) -> system_error EIO s
        (Format.sprintf "%s/%s" mes (Unix.error_message error))


(* --------------------- *)
(* Structures de données *)
(* --------------------- *)

(* Le type [partition] contient les principales valeurs caractérisant
   une partition. *)
type partition =
    { part_name: string;
      (* le nom du fichier sur le disque *)
      part_fd: Unix.file_descr;
      (* le descripteur de fichier correspondant *)
      block_size: int;
      (* la taille des blocs (en octects) *)
      block_nb: int;
      (* le nombre de blocs *)
      mutable part_alive: bool;
      (* un booléen indiquant si la partition est activée, c'est a
         dire si le champ part_fd est toujours
         ouvert *)
      block_word_size: int;
      (* le nombre de mots par bloc (entiers de 4 octects dans un
         bloc) *)
      blocktbl_size: int;
      (* le nombre maximal de blocs dans un fichier, i.e. la taille en
         mots de la table des blocs d'un inode *)
      max_file_size: int;
      (* la taille maximale d'un fichier, calculee a partir des
         proprietes de la partition *)
      blocktbl_offset: int;
      (* la position de la table des blocs dans un inode, en
         octects.  *)
      inode_nb: int;
      (* le nombre d'inodes dans la partition *)
      mutable last_free_inode: int;
      (* le numéro du dernier inode libéré *)
      mutable free_block_list: int;
      (* le numéro du bloc index de la liste de blocs libres *)
      mutable free_block_nb: int;
      (* la position dans ce bloc d'index *)
      fstype: string;
      (* le type de la partition *)
      root_inode: int;
      (* le numéro de l'inode du répertoire racine *)
    }

(* Définition d'un i-node *)
type inode =
    { stats: stats;
      blocktbl: int array;
      partition: partition; }

and file_kind = S_REG | S_DIR

and stats =
    { st_dev: int;
      st_ino: int;
      mutable st_kind: file_kind;
      mutable st_nlink: int;
      mutable st_size: int; }

(* caractéristique d'un répertoire *)
let dirent_size = 32
let filename_max_size = 28

(* --------------------- *)
(* Fonctions utilitaires *)
(* --------------------- *)

exception Implementation of string

let implementation_error s = raise (Implementation s)

let rec really_read fd buffer start length =
  if length <= 0 then () else
  match Unix.read fd buffer start length with
  | 0 -> raise End_of_file
  | r -> really_read fd buffer (start + r) (length - r);;

(* [read_int] permet de lire des entiers codés sur 4 caractères. *)
let read_int =
  let get_int8 s pos = int_of_char s.[pos] in
  fun s pos ->
    let c4 = get_int8 s pos in
    let c3 = get_int8 s (pos+1) in
    let c2 = get_int8 s (pos+2) in
    let c1 = get_int8 s (pos+3) in
    c1 + (c2 lsl 8) + (c3 lsl 16) + (c4 lsl 24)


let print_partition part =
  Format.printf "Partition %s (%s)@\n  @["
    part.part_name
    (if part.part_alive then "alive" else "dead");
  Format.printf "block_size: %i@\n" part.block_size;
  Format.printf "block_nb: %i@\n" part.block_nb;
  Format.printf "block_word_size: %i@\n" part.block_word_size;
  Format.printf "i-nodes@\n  @[";
  Format.printf "blocktbl_size: %i@\n" part.blocktbl_size;
  Format.printf "max_file_size: %i@\n" part.max_file_size;
  Format.printf "blocktbl_offset: %i@]@\n" part.blocktbl_offset;
  Format.printf "supernode@\n  @[";
  Format.printf "inode_nb: %i@\n" part.inode_nb;
  Format.printf "last_free_inode: %i@\n" part.last_free_inode;
  Format.printf "free_block_list: %i@\n" part.free_block_list;
  Format.printf "free_block_nb: %i@\n" part.free_block_nb;
  Format.printf "fstype: %s@\n" part.fstype;
  Format.printf "root_inode: %i@]@]@\n" part.root_inode

let print_block part bid b =
  Format.printf "Block %i:@\n  @[" bid;
  for i = 0 to (part.block_size / 4) - 1 do
    Format.printf "%i@ " (read_int b (i*4))
  done;
  Format.printf "@\n@?"


(* ------------------------- *)
(* Début du code à completer *)
(* ------------------------- *)

let open_partition name =
  let  fd = handle_unix (Unix.openfile name [ Unix.O_RDWR ]) 0 in

  (* read the last 4 bytes *)
  ignore (Unix.lseek fd  (-4) Unix.SEEK_END);
  let buffer = String.create 4 in
  really_read fd buffer 0 4;
  let block_size = read_int buffer 0 in

  let size = (Unix.fstat fd).Unix.st_size in
  assert ((size - 4) mod block_size = 0);

  (* compute some values *)
  let blocktbl_offset = 12 in
  let block_nb =  (size - 4) / block_size in
  let blocktbl_size = (block_size - blocktbl_offset) / 4 in
  let max_file_size = block_size * blocktbl_size in
  let block_word_size = block_size / 4  in

  (* read the superblock *)
  ignore (Unix.lseek fd  0 Unix.SEEK_SET);
  let buffer = String.create block_size  in
  really_read fd buffer 0 block_size;
  let inode_nb = read_int buffer 0 in
  let last_free_inode =  read_int buffer 4 in
  let free_block_list = read_int buffer 8 in
  let free_block_nb = read_int buffer 12 in
  let root_inode = read_int buffer 16 in
  assert (root_inode = 1);
  let fstype = String.sub buffer (block_size - 4) 4 in
  assert (fstype = "sfs0");

  (* generate the partition value *)
  {
    part_name = name;
    part_fd = fd;
    block_size = block_size;
    block_nb = block_nb;
    part_alive = true;

    block_word_size = block_word_size;

    blocktbl_offset = blocktbl_offset;
    blocktbl_size = blocktbl_size;
    max_file_size = max_file_size;

    inode_nb = inode_nb;
    last_free_inode = last_free_inode;
    free_block_list = free_block_list;
    free_block_nb = free_block_nb;
    fstype = fstype;
    root_inode = root_inode;
  }


let lseek_block p n =
  assert (n < p.block_nb);
  assert (p.part_alive);
  ignore (handle_unix (Unix.lseek p.part_fd (n * p.block_size)) SEEK_SET)

let read_block p n =
  if not p.part_alive then system_error EIO "read_block: not alive" p.part_name;
  let buffer = String.create p.block_size in
  try lseek_block p n;
      really_read p.part_fd buffer 0 p.block_size;
      buffer
  with Unix.Unix_error (_,_,_) | End_of_file ->
    system_error EIO "readblock: Unix_error" p.part_name

let read_inode p n =
  let buffer = read_block p n in
  let kind =
    match (read_int buffer 0) with
    | 0 -> S_REG
    | 1 -> S_DIR
    | _ -> system_error EIO "read_inode" "kind" in
  let nlink = read_int buffer 4 in
  let size = read_int buffer 8 in
  if size > p.max_file_size then system_error EIO "read_inode" "size";
  let init_block i = (read_int buffer (i*4 + p.blocktbl_offset)) in
  let blocktbl = Array.init (p.blocktbl_size) init_block in
  { stats =  { st_dev = 0;
               st_ino = n ;
               st_kind = kind ;
               st_nlink = nlink;
               st_size = size; };
    blocktbl = blocktbl;
    partition = p;
  }

type file_descr = { inode : inode; mutable pos : int; mutable closed : bool }

let open_inode inode =
  { inode = inode;
    pos = 0;
    closed = false; }

let read desc buffer from len =
  let p = desc.inode.partition in
  if desc.closed then
    system_error EINVAL "read" "closed descr";
  if desc.inode.stats.st_size > p.max_file_size then
    system_error EIO "read" "Inconsistent file system";
  if len <= 0 then
    system_error EINVAL "read" "negative length";
  if from + len > String.length buffer then
    system_error EINVAL "read" "too small buffer";
  let len = min len (desc.inode.stats.st_size - desc.pos) in
  let rec read start len ret =
    if len > 0 then
      let block = desc.pos / p.block_size in
      let pos' = desc.pos mod p.block_size in
      let len' = min len (p.block_size - pos') in
      let bid = desc.inode.blocktbl.(block) in
      if bid > 0 then
        let b = read_block p bid in
        String.blit b pos' buffer start len'
      else
        String.fill buffer start len' '\000';
      desc.pos <- desc.pos + len';
      read (start + len') (len - len') (ret + len')
    else ret
  in
  read from len 0


let rec read_dirent desc =
  let dirent = String.create dirent_size in
  let nread = read desc dirent 0 dirent_size in
  if nread = 0 then raise End_of_file ;
  if dirent.[0] = '\000' then read_dirent desc
  else
    let inode = read_int dirent filename_max_size in
    dirent.[filename_max_size] <- '\000';
    inode, String.sub dirent 0 (String.index dirent '\000')

let rec print_directory inode dirname =
  let p = inode.partition in
  let fd = open_inode inode in
  try
    while true do
      let (i, name) = read_dirent fd in
      let filename = Filename.concat dirname name in
      Format.printf "%i %s@\n" i filename;
      let inode = read_inode p i in
      match inode.stats.st_kind with
      | S_REG -> ()
      | S_DIR ->
          if name <> "." && name <> ".." then
            print_directory inode filename
    done
  with End_of_file -> ()


type block_type =
  | Inode of int    (* Inode, nombre de liens durs observés *)
  | Block of bool   (* Block, utilisé ou pas *)
  | IndexBlock      (* Block d'index de la liste des blocs libres *)
  | Supernode       (* Supernode du système *)
  | FreeBlock       (* Block libre confirmé *)

let string_of_block t =
  match t with
  | Inode n -> Format.sprintf "inode[%d]" n
  | Block b -> Format.sprintf "block[%b]" b
  | IndexBlock -> "indexblock"
  | Supernode -> "supernode"
  | FreeBlock -> "freeblock"

let fsck p =
  (* initialisation de la table des blocs *)
  let t = Array.create p.block_nb (Block false) in
  t.(0) <- Supernode;
  for i = 1 to p.inode_nb do
    t.(i) <- Inode 0
  done;

  (* parcours de répertoires *)
  Format.printf "Scanning directory tree@\n";
  let rec fsck_inode inode =
    Format.printf ".";
    match t.(inode) with
    | Inode 0 -> begin
        t.(inode) <- Inode 1;
        let i = read_inode p inode in
        for n = 0 to p.blocktbl_size - 1 do
          let b = i.blocktbl.(n) in
          begin match t.(b) with
          | Block false ->
              if p.block_size * n > i.stats.st_size then
                Format.printf "Inode %d uses a block after its size@\n" inode
              else
                t.(b) <- Block true
          | Supernode -> ()
          | _ ->
              Format.printf "Inode %d uses block %d %s@\n" inode b
                (string_of_block t.(b))
          end
        done;
        begin match i.stats.st_kind with
        | S_REG -> ()
        | S_DIR ->
            let fd = open_inode i in
            begin try
              while true do
                let (inode, name) = read_dirent fd in
                fsck_inode inode
              done
            with End_of_file -> ()
            end
        end
      end
    | Inode n -> t.(inode) <- Inode (n+1)
    | _ -> Format.printf "Inode %d is not an inode !@\n" inode
  in
  fsck_inode p.root_inode;
  Format.printf "@\n";

  (* vérification des inodes *)
  for n = 1 to p.inode_nb do
    let i = read_inode p n in
    match t.(n) with
    | Inode x ->
        if x <> i.stats.st_nlink then
          Format.printf "%d should have %i links !@\n" n x
    | _ -> Format.printf "%d should be an inode !@\n" n
  done;

  (* parcours de la liste des blocs libres *)
  Format.printf "Scanning free block list@\n";
  let rec iter block nb =
    if block <> 0 then
      match t.(block) with
      | Block false ->
          Format.printf "(%d)" nb;
          t.(block) <- IndexBlock;
          let buffer = read_block p block in
          let next = read_int buffer 0 in

          for i = 1 to nb - 1 do
            let b = read_int buffer (i * 4) in
            match t.(b) with
            | Block false -> t.(b) <- FreeBlock
            | _ ->
                 Format.printf "Free Block %d has type %s@\n"
                   b (string_of_block t.(b))
          done;
          iter next (p.block_word_size - 1)
      | _ ->
          Format.printf "Block %d is not an index block !@\n" block
  in
  iter p.free_block_list ( (p.free_block_nb - 1) mod p.block_word_size);
  Format.printf "@\n";

  (* vérifier que tous les blocs ont été traités *)
  for n = p.inode_nb + 1 to p.block_nb - 1 do
    match t.(n) with
    | Block false -> Format.printf "Forgotten block %d\n" n
    | _ -> ()
  done;

  Format.printf "Scan done@\n@?";
  ()

let () =
  let usage =  Format.sprintf "Usage: %s file" Sys.argv.(0) in
  if not (Array.length Sys.argv = 2) then begin
    Format.eprintf "%s@." usage;
    exit 1
  end;
  let part = open_partition Sys.argv.(1) in
  print_partition part;
  Format.printf "--> ls -Ri /@\n";
  print_directory (read_inode part part.root_inode) "/";
  Format.printf "--> fsck@\n";
  fsck part
